=== modified file 'NEWS'
--- NEWS	2008-07-16 18:14:23 +0000
+++ NEWS	2008-07-16 23:23:13 +0000
@@ -73,6 +73,9 @@
       still in ``.bzr.log`` and can be shown with ``-Derror``.
       (Martin Pool, #240161)
 
+    * Fetch from a stacked branch copies all required data.
+      (Aaron Bentley, #248506)
+
     * Handle urls such as ftp://user@host.com@www.host.com where the user
       name contains an @.
       (Neil Martinsen-Burrell, #228058)

=== modified file 'bzrlib/repository.py'
--- bzrlib/repository.py	2008-07-16 07:30:18 +0000
+++ bzrlib/repository.py	2008-07-16 14:23:05 +0000
@@ -2709,6 +2709,11 @@
     @needs_write_lock
     def fetch(self, revision_id=None, pb=None, find_ghosts=False):
         """See InterRepository.fetch()."""
+        if len(self.source._fallback_repositories) > 0:
+            from bzrlib.fetch import KnitRepoFetcher
+            fetcher = KnitRepoFetcher(self.target, self.source, revision_id,
+                                      pb, find_ghosts)
+            return fetcher.count_copied, fetcher.failed_revisions
         from bzrlib.repofmt.pack_repo import Packer
         mutter("Using fetch logic to copy between %s(%s) and %s(%s)",
                self.source, self.source._format, self.target, self.target._format)

=== modified file 'bzrlib/tests/branch_implementations/test_stacking.py'
--- bzrlib/tests/branch_implementations/test_stacking.py	2008-07-16 07:30:18 +0000
+++ bzrlib/tests/branch_implementations/test_stacking.py	2008-07-16 22:58:50 +0000
@@ -172,15 +172,8 @@
                 errors.UnstackableRepositoryFormat):
             # not a testable combination.
             return
-        try:
-            cloned_unstacked_bzrdir = stacked_bzrdir.clone('cloned-unstacked',
-                preserve_stacking=False)
-        except errors.NoSuchRevision:
-            raise KnownFailure(
-                'Pack-to-pack fetch does not handle stacking properly.'
-                ' (#248506)')
-        else:
-            self.fail('Expected a failure due to broken fetching.')
+        cloned_unstacked_bzrdir = stacked_bzrdir.clone('cloned-unstacked',
+            preserve_stacking=False)
         unstacked_branch = cloned_unstacked_bzrdir.open_branch()
         self.assertRaises((errors.NotStacked, errors.UnstackableBranchFormat),
                           unstacked_branch.get_stacked_on_url)
@@ -215,3 +208,26 @@
             self.assertEqual('../stack-on', target.get_stacked_on_url())
         except errors.UnstackableBranchFormat:
             pass
+
+    def prepare_stacked_on_fetch(self):
+        stack_on = self.make_branch_and_tree('stack-on')
+        stack_on.commit('first commit', rev_id='rev1')
+        try:
+            stacked_dir = stack_on.bzrdir.sprout('stacked', stacked=True)
+        except (errors.UnstackableRepositoryFormat,
+                errors.UnstackableBranchFormat):
+            raise TestNotApplicable('Format does not support stacking.')
+        unstacked = self.make_repository('unstacked')
+        return stacked_dir.open_workingtree(), unstacked
+
+    def test_fetch_copies_from_stacked_on(self):
+        stacked, unstacked = self.prepare_stacked_on_fetch()
+        unstacked.fetch(stacked.branch.repository, 'rev1')
+        unstacked.get_revision('rev1')
+
+    def test_fetch_copies_from_stacked_on_and_stacked(self):
+        stacked, unstacked = self.prepare_stacked_on_fetch()
+        stacked.commit('second commit', rev_id='rev2')
+        unstacked.fetch(stacked.branch.repository, 'rev2')
+        unstacked.get_revision('rev1')
+        unstacked.get_revision('rev2')

